'''
Created on 2016-03-26

@author: bdillman
'''

import os
import shutil

testDest = "testBox"
testSrc  = "testApp"

def copyTestApp():
    shutil.copytree(testSrc, testDest)

def deleteTestApp():
    try:
        shutil.rmtree(testDest)
    except FileNotFoundError:
        pass

def mergeTree(srcDir, destDir):
    if not os.path.isdir(srcDir):
        shutil.copy(srcDir, destDir)
        return
    for src in os.listdir(srcDir):
        srcFile  = "{0}/{1}".format(srcDir,  src)
        destFile = "{0}/{1}".format(destDir, src)
        if(os.path.isdir(srcFile)):
            mergeTree(srcFile, destFile)
        else:
            shutil.copy(srcFile, destFile)
    
def copyFashionStuff():
    mergeTree("../testFashion/fashion/model",    "./fashion/model")
    mergeTree("../testFashion/fashion/template", "./fashion/template")
    mergeTree("../testFashion/fashion/xform",    "./fashion/xform")



'''
Created on 2016-05-01

Copyright (c) 2016 Bradford Dillman

@author: bdillman

Adapt mako templates for use by fashion application.
'''

import logging
import os.path
import pathlib
import shutil

from mako.lookup import TemplateLookup



#
# The list of directories used to lookup template files.
#
templateLookup = None

#
# Where to keep backups of the generated files.
# This is used to check for modification before overwrite.
#
mirrorDirectory = None

#
# The root project directory. All generation must be under this.
#
projectDirectory = None

#
# True if we force an overwrite (default to False).
#
overwrite = False


def setDirectories(projDir=None, mirrorDir=None, templateDirs=None, moduleDir=None):
    '''Set the directories where templates are found, for later lookup.'''
    global templateLookup
    global mirrorDirectory
    global projectDirectory
    projectDirectory = projDir
    mirrorDirectory = mirrorDir
    logging.debug(str(templateDirs))
    templateLookup = TemplateLookup(templateDirs, module_directory=moduleDir)

def generateFile(model=None, templateFile=None, targetFile=None, **kwargs):
    '''Generate a file using a model and a template.'''
    logging.debug("generateFile(model={0}, templateFile={1}, targetFile={2}".format(model, templateFile, targetFile))
    logging.debug("**kwargs={0}".format(kwargs))
    
    global overwrite
    
    # get the relative dir for target
    relFile = os.path.relpath(targetFile, str(projectDirectory))
    mirrorFile = os.path.join(mirrorDirectory, relFile)
    
    # check for file overwrite
    if os.path.exists(targetFile):
        if os.path.exists(mirrorFile):
            targetTime = os.path.getmtime(targetFile)
            mirrorTime = os.path.getmtime(mirrorFile)
            if targetTime > mirrorTime:
                if overwrite:
                    logging.warn("target modified - overwrite: {0}".format(relFile))
                else:
                    logging.warn("target modified - skipping: {0}".format(relFile))
                    return
       
    
    mytemplate = templateLookup.get_template(templateFile)
    pathlib.Path(os.path.dirname(targetFile)).mkdir(parents=True, exist_ok=True)
    with open(targetFile, 'w') as f:
        f.write(mytemplate.render(**model))

    # copy to mirror directory
    pathlib.Path(os.path.dirname(mirrorFile)).mkdir(parents=True, exist_ok=True)
    shutil.copy2(targetFile, mirrorFile)
    
def createDefaultFile(model=None, templateFile=None, targetFile=None, **kwargs):
    '''Create a default model, template, xform or other file.'''
    mytemplate = templateLookup.get_template(templateFile)
    pathlib.Path(os.path.dirname(targetFile)).mkdir(parents=True, exist_ok=True)
    with open(targetFile, 'w') as f:
        f.write(mytemplate.render(**model))
    


'''
Created on 2016-04-17

Copyright (c) 2016 Bradford Dillman

@author: bdillman

Manage files using metadata stored in a database.
'''
import datetime
import logging
import os.path

import peewee
import playhouse.sqlite_ext
import ruamel.yaml



database_proxy = peewee.Proxy()
database = None

def init_db(filename):
    '''Initialize a connection to the database.'''
    logging.debug("init_db({})".format(filename))
    global database
    global database_proxy
    database = peewee.SqliteDatabase(filename)
    database_proxy.initialize(database)
    database.connect()
    return True

def create_db():
    '''Create a new database schema from an existing database connection.'''
    database.create_tables([FashionFile], safe=True)



class BaseModel(peewee.Model):
    '''Just for architectures' sake.'''
    class Meta:
        database = database_proxy




class FashionFile(BaseModel):
    '''Database record with metadata about a file in the filesystem.'''

    id = playhouse.sqlite_ext.PrimaryKeyAutoIncrementField(index=True,
                                                           unique=True,
                                                           null=False,
                                                           primary_key=True)
    
    # role target=1, template=2, model=3, xform=4
    role = peewee.IntegerField(index=True, unique=False, null=False)
    
    # (relative?) filename
    filename = peewee.CharField(index=True, unique=True, null=True)
    
    # last know timestamp of file
    timestamp = peewee.DateTimeField(index=True, unique=False, null=True)
    
    # model kind
    kind = peewee.CharField(index=True, unique=False, null=True)     
    
    # file format
    fileFormat = peewee.CharField(index=True, unique=False, null=True)
    
    # content of anonymous file
    content = peewee.BlobField(index=False, unique=False, null=True)
    
    # last time this record was modified
    lastModTime = peewee.DateTimeField(default=datetime.datetime.now)
    
    def loadFile(self):
        '''
        Load the file from the filesystem.
        Returns True for success, False for failure.
        '''
        if self.role == 3: # model
            if self.fileFormat == 'yaml': 
                if self.filename == None:
                    self.model = ruamel.yaml.load(self.content)
                    return self.model
                else:
                    with open(self.filename, 'r') as stream:
                        self.model = ruamel.yaml.load(stream)
                        return self.model
    
    def exists(self):
        '''Returns True iff filename exists on the filesystem.'''
        return os.path.exists(self.filename)
    
    def saveFile(self):
        '''Save self.model to filename.'''
        if self.content == None:
            raise("nothing to save")
        if self.filename == None:
            raise("no filename")
        with open(self.filename, 'w') as file:
            logging.debug("FashionFile.save({})".format(self.filename))
            file.write(self.content)
    


def importFile(**kwargs):
    '''Import a file with metadata.'''
    logging.debug("importFile({0})".format(kwargs))
    return FashionFile.get_or_create(**kwargs) 

def createModel(model, kind, fileFormat='yaml', filename=None):
    '''Create a new model with metadata.'''
    if fileFormat == 'yaml':
        content = ruamel.yaml.dump(model, default_flow_style = False)
    file = FashionFile.create(role=3, kind=kind, content=content, filename=filename, fileFormat=fileFormat)
    if filename != None:
        file.saveFile()
    return file

def getProjectModel():
    '''Get the project model for this project.'''
    pm = ruamel.yaml.load(FashionFile.get(kind='fashion_project').content)
    logging.debug("pm={0}".format(pm))
    return pm

def getModelFiles(kind):
    '''Get all models of the given kind.'''
    return FashionFile.select().where(FashionFile.role==3, FashionFile.kind==kind)
    
def clean():
    '''Delete model metadata for all but the project model.'''
    q = FashionFile.delete().where((FashionFile.role==3) & (FashionFile.kind != 'fashion_project'))
    q.execute()


'''
Created on 2016-04-29

Copyright (c) 2016 Bradford Dillman

@author: bdillman

Utilities used by user xforms.
'''

import logging
import os.path

from . import fashionPortfolio

#
# Name of current executing xform
#
current_xform_name = ""

#
# Set of current output kinds
#
current_output_kinds = set()

class cd:
    '''Context manager for changing the current working directory'''
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        logging.debug("cd {0} -> {1}".format(self.savedPath, self.newPath))
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        logging.debug("cd {1} -> {0}".format(self.savedPath, self.newPath))
        os.chdir(self.savedPath)



def readModels(modelFashionFiles, flatten=False):
    '''Read a list of models.'''
    modelList = [f.loadFile() for f in modelFashionFiles]
    if flatten:
        models = [item for sublist in modelList for item in sublist]
    else:
        models = modelList
    return models
    
def writeModel(model=None, kind=None, fileFormat='yaml', filename=None):
    '''Write a single model.'''
    if kind not in current_output_kinds:
        logging.warn("xform {0} didn't declare output kind: {1}".format(current_xform_name, current_output_kinds))
    fashionPortfolio.createModel(model, kind, fileFormat, filename)

def generate(model=None, templateFile=None, targetFile=None):
    '''Write a generation model for a single file.'''
    genModel = { 'model'       : model,
                 'templateFile': templateFile,
                 'targetFile'  : targetFile }    
    writeModel(genModel, 'fashion_gen')



'''
Created on 2016-03-25

@author: bdillman
'''
import unittest

import os
import logging

import testUtils
import fashionCmds


class Test(unittest.TestCase):
    """
    Test commands for initializing and deleting a project.
    """
    
    def setUp(self):
        testUtils.deleteTestApp()
        testUtils.copyTestApp()
        os.chdir(testUtils.testDest)
        logging.basicConfig(level=logging.DEBUG)

    def tearDown(self):
        os.chdir("..")
        testUtils.deleteTestApp()
        
    def runCmd(self, args):
        args = ["-y"] + args
        fashionCmds.main(args)

    def testInit(self):
        self.runCmd(["init"])

    def testInit2(self):
        self.runCmd(["init"])
        self.runCmd(["init"])
 
    def testKill(self):
        self.runCmd(["init"])
        self.runCmd(["kill"])
 
    def testKill2(self):
        self.runCmd(["kill"])

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on 2016-04-29

Copyright (c) 2016 Bradford Dillman

@author: bdillman

Project class describes a fashion user project.
'''

import os
import shutil
import logging
import pathlib

import peewee

from . import fashionPortfolio
from . import library
from . import templates



#
# Get the FASHION_HOME directory.
#
FASHION_HOME = pathlib.Path(os.path.abspath(os.path.dirname(__file__)))



class Project:
	'''Represents a fashion user project.'''

	def __init__(self, projDir):
		'''
		Initialize a project mapped to the given directory, whether or not 
		the project or the directory actually exist.
		
		:param str projDir: where the project is located. 
		'''
		self.projectPath   = pathlib.Path(projDir)
		self.fashionPath   = self.projectPath / 'fashion'
		self.fashionDbPath = self.fashionPath / 'fashion.db'

	def exists(self):
		'''Check if this project exists.'''
		return self.fashionDbPath.exists()

	def findRelativePath(self, filename):
		'''
		Get a path or filename relative to the project path.
		
		:param str filename: input filename.
		'''
		return os.path.relpath(filename, str(self.projectPath))
	
	def create(self):
		'''Create a new fashion project.'''
		try:
			# create the project model
			projectModel = { 'projectPath'   : str(self.projectPath),
							 'fashionPath'   : str(self.fashionPath),
							 'fashionDbPath' : str(self.fashionDbPath),
							 'tmplMods'      : str(self.fashionPath / 'tmplMods'),
							 'bakPath'       : str(self.fashionPath / 'bak'),
							 'mirrorPath'    : str(self.fashionPath / 'mirror'),
							 'libraries'     : [str(self.fashionPath / 'library.yaml'),
											    str(FASHION_HOME     / 'library.yaml') ] }
			logging.debug("Creating fashion project: {0}".format(projectModel['projectPath']))
			
			tmplMods     = str(self.fashionPath / 'tmplMods')
			modelPath    = str(self.fashionPath / 'model')
			templatePath = str(self.fashionPath / 'template')
			xformPath    = str(self.fashionPath / 'xform')
			
			# create the ./fashion subdirs
			os.mkdir(projectModel['fashionPath'])
			os.mkdir(projectModel['bakPath'])
			os.mkdir(projectModel['mirrorPath'])
			os.mkdir(tmplMods)
			os.mkdir(modelPath)
			os.mkdir(templatePath)
			os.mkdir(xformPath)
			
			# create the database connection
			try:
				self.init_db()
			except peewee.OperationalError:
				# create the database schema
				fashionPortfolio.create_db()
				filename=os.path.join(projectModel['fashionPath'],"fashion.yaml")
				fashionPortfolio.createModel(projectModel, 'fashion_project', filename=filename)
				
			# default library
			lib = library.Library(projectModel['libraries'][0])
			lib.addGlob(glob="./model/", recursive=False,
					    role=3, kind='fashion_unknown', fileFormat='yaml')
			lib.addGlob(glob="./xform/", recursive=False,
					    role=4, fileFormat='python3')
			lib.addGlob(glob="./template/", recursive=False,
					    role=2, fileFormat='mako')
			lib.addGlob(glob="./xform/**/*.py", recursive=True,
					    role=4, fileFormat='python3')
			lib.save()
			
			self.projectModel = projectModel
			
			self.importModels()
				
		except FileExistsError:
			logging.warning("Can't create fashion project: {0} already exists".format(str(self.projectPath)))

	def importModels(self):
		'''Import all models from all libraries.'''
		for libFilename in self.projectModel['libraries']:
			l = library.Library(libFilename)
			l.load()
			l.importModels()
			
	def loadXforms(self):
		'''Import all xforms from all libraries.'''
		xfList = []
		for libFilename in self.projectModel['libraries']:
			l = library.Library(libFilename)
			l.load()
			xfList.extend(l.loadXforms())
		return xfList
	
	def loadTemplates(self):
		'''Load all template directories.'''
		dirs = []
		for libFilename in self.projectModel['libraries']:
			l = library.Library(libFilename)
			l.load()
			dirs.extend(l.getDirectories(2))
		templates.setDirectories(self.projectModel['projectPath'], self.projectModel['mirrorPath'], dirs, self.projectModel['tmplMods'])
		
	def getLocalLibrary(self):
		l = library.Library(self.projectModel['libraries'][0])
		l.load()
		return l
		
	def getLocalModelDir(self):
		return self.getLocalLibrary().getDirectories(3)[0]
		
	def getLocalTemplateDir(self):
		return self.getLocalLibrary().getDirectories(2)[0]
		
	def getLocalXformDir(self):
		return self.getLocalLibrary().getDirectories(4)[0]
	
	def destroy(self):
		'''Destroy an existing fashion project.''' 
		if self.exists():
			logging.debug("Destroying fashion project: {0}".format(str(self.projectPath)))
			shutil.rmtree(str(self.fashionPath))
		else:
			logging.debug("Can't destroy fashion project: {0} doesn't exist".format(str(self.projectPath)))
			
	def init_db(self):
		'''Initialize the fashion database.'''
		if fashionPortfolio.database == None:
			filename = str(self.fashionDbPath)
			fashionPortfolio.init_db(filename)
			
		self.projectModel = fashionPortfolio.getProjectModel()
		


def findProject(startDir):
	'''
	Find the root directory of a project by searching recursively upwards.
	
	:param str startDir: where to begin searching for project.
	:returns: the Project object, if found; otherwise None.
	:rtype: Project
	'''
	p = Project(startDir)
	while True:
		if p.exists():
			logging.debug("Found project: {0}".format(str(p.projectPath)))
			return p
		else:
			parent = p.projectPath.parent
			if parent == p.projectPath:
				break
			p = Project(parent)
	logging.debug("No project found: {0}".format(startDir))
	return None




'''
Created on 2016-03-26

@author: bdillman
'''
import unittest
import os
import logging

import testUtils
import project

class Test(unittest.TestCase):


    def setUp(self):
        testUtils.deleteTestApp()
        testUtils.copyTestApp()
        os.chdir(testUtils.testDest)
        logging.basicConfig(level=logging.DEBUG)

    def tearDown(self):
        os.chdir("..")
        testUtils.deleteTestApp()


    def testProjectExists(self):
        p = project.Project(os.getcwd())
        self.assertFalse(p.exists(), "found project in cwd")
        self.assertEqual(str(p.projectPath), os.getcwd())
        p.create()
        self.assertTrue(p.exists(), "found project in cwd")
        

    def testFindProjectInCwd(self):
        p = project.findProject(os.getcwd())
        self.assertIsNone(p, "found project in cwd")
        p = project.Project(os.getcwd())
        p.create()
        p = project.findProject(os.getcwd())
        self.assertIsNotNone(p, "found project in cwd")
        self.assertEqual(str(p.projectPath), os.getcwd())

    def testFindProjectInParent(self):
        p = project.findProject(os.getcwd())
        self.assertIsNone(p, "found project in cwd")
        p = project.Project(os.getcwd())
        p.create()
        os.chdir("bar")
        p = project.findProject(os.getcwd())
        self.assertIsNotNone(p, "found project in cwd")
        os.chdir("..")
        self.assertEqual(str(p.projectPath), os.getcwd())
        
        
    def testGetXformMgr(self):
        p = project.Project(os.getcwd())
        xm = p.getXformManager()
        self.assertIsNotNone(xm, "found xformMgr in project")
        

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on 2016-03-26

@author: bdillman
'''
import unittest
import os
import logging

import testUtils

import xforms
import project

class Test(unittest.TestCase):


    def setUp(self):
        testUtils.deleteTestApp()
        testUtils.copyTestApp()
        os.chdir(testUtils.testDest)
        p = project.Project(os.getcwd())
        p.create()
        testUtils.copyFashionStuff()
        logging.basicConfig(level=logging.DEBUG)

    def tearDown(self):
        os.chdir("..")
        testUtils.deleteTestApp()


    def testFindXformLocal(self):
        xfm = xforms.XformManager(["/bad/path", "fashion/xform"])
        self.assertIsNotNone(xfm, "found XformManager")
        xf = xfm.find("myXform")
        self.assertIsNotNone(xf, "found myXform")


    def testFindXformFashion(self):
        xfm = xforms.XformManager(["/bad/path", "fashion/xform", "../../xform"])
        self.assertIsNotNone(xfm, "found XformManager")
        xf = xfm.find("gen")
        self.assertIsNotNone(xf, "found myXform")


    def testXformExists(self):
        xf = xforms.Xform("fashion/xform/myXform.py")
        self.assertTrue(xf.exists(), "exists")
        xf = xforms.Xform("fashion/xform/notMyXform.py")
        self.assertFalse(xf.exists(), "exists")

    def testLoadXformLocal(self):
        xfm = xforms.XformManager(["/bad/path", "fashion/xform", "../../xform"])
        self.assertIsNotNone(xfm, "found XformManager")
        xf = xfm.find("myXform")
        self.assertIsNotNone(xf, "found myXform")
        xf.load()
        self.assertIsNotNone(xf.mod, "found python module")


    def testLoadXformFashion(self):
        xfm = xforms.XformManager(["/bad/path", "fashion/xform", "../../xform"])
        self.assertIsNotNone(xfm, "found XformManager")
        xf = xfm.find("gen")
        self.assertIsNotNone(xf, "found myXform")
        xf.load()
        self.assertIsNotNone(xf.mod, "found python module")


    def testFindAll(self):
        xfm = xforms.XformManager(["/bad/path", "fashion/xform", "../../xform"])
        self.assertIsNotNone(xfm, "found XformManager")
        xfList = xfm.findAll()
        self.assertIsNotNone(xfList, "found myXform")
        xfNames = [xf.modName for xf in xfList]
        self.assertIn("gen", xfNames)
        self.assertIn("myXform", xfNames)


    def testLoadAll(self):
        xfm = xforms.XformManager(["/bad/path", "fashion/xform", "../../xform"])
        self.assertIsNotNone(xfm, "found XformManager")
        xfm.loadAll()
        self.assertIsNotNone(xfm.xforms, "found myXform")
        self.assertEqual(len(xfm.xforms), 2, "number of xforms loaded")
        xfNames = [xf.modName for xf in xfm.xforms]
        self.assertIn("gen", xfNames)
        self.assertIn("myXform", xfNames)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on 2016-04-29

Copyright (c) 2016 Bradford Dillman

@author: bdillman

xform which generates code from models and templates.
'''

import logging

from fashion import templates

logging.debug("fashion_gen loaded")

def outputKinds():
    '''Outputs generated by this xform'''
    return [ ]

def police(fashion_gen):
    '''Validate models before xform.'''
    pass

def xform(fashion_gen):
    '''Perform source generation from template.'''
    logging.debug("fashion_gen xform running")
    for f in fashion_gen:
        m = f.loadFile()
        templates.generateFile(**m)



'''
Created on 2016-03-25

@author: bdillman
'''
import unittest

import os
import logging

import testUtils
import fashionCmds
import project


class Test(unittest.TestCase):
    """
    Test commands.
    """

    def setUp(self):
        testUtils.deleteTestApp()
        testUtils.copyTestApp()
        os.chdir(testUtils.testDest)
        logging.basicConfig(level=logging.DEBUG)
        self.runCmd(["init"])
        testUtils.copyFashionStuff()

    def tearDown(self):
        os.chdir("..")
        testUtils.deleteTestApp()
        
    def runCmd(self, args):
        args = ["-y"] + args
        fashionCmds.main(args)


#     def testVersion(self):
#         self.runCmd(["version"])
        
    @unittest.skip("not yet")
    def testXformGen(self):
        p = project.Project(os.getcwd())
        p.init_db()
        tgtMgr = p.getTargetManager()
        tgtMgr.createTarget("test.txt", "hello_world.tlt", "hello_world.yaml")
        self.runCmd(["xform", "gen"])

    def testXformGreetWorlds(self):
        p = project.Project(os.getcwd())
        p.init_db()
        tgtMgr = p.getTargetManager()
        tgtMgr.createTarget("test.txt", "hello_world.tlt", "hello_world.yaml")
        self.runCmd(["xform", "greetWorlds"])
        
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testVersion']
    unittest.main()


'''
${name} xform.
'''

import logging

from fashion.xformUtil import readModels
from fashion.xformUtil import generate

def outputKinds():
    '''Outputs generated by this xform'''
    return [ 'fashion_gen' ]

def police():
    '''Validate input combinations.'''
    pass

def xform():
    '''Generate many source files from 2 input models.'''
    logging.debug("${name} xform running")
    
    model = {  }
    
    generate(model, "${template}", "${target}")



'''
Created on 2016-04-29

Copyright (c) 2016 Bradford Dillman

@author: bdillman

Describe and manipulate collections of metadata for models, templates and 
xforms.
'''

import fnmatch
import glob
import logging
import os.path

import ruamel.yaml

from . import fashionPortfolio
from . import xforms
from . import xformUtil



class Library(object):
    '''A collection of globs describing metadata for models, templates and xforms.'''

    def __init__(self, filename):
        '''Constructor'''
        self.filename = filename
        self.globs = []
        
    def matchGlobs(self, filename):
        return [g for g in self.globs
                  for path in glob.glob(g['glob'], recursive=g['recursive'])
                  if fnmatch.fnmatch(filename, path)]
        
    def addFile(self, filename, role, kind='fashion_unknown', fileFormat='yaml'):
        matchedGlobs = self.matchGlobs(filename)
        for g in matchedGlobs:
            raise("conflicting glob: {0}".format(g))
        else:
            self.addGlob(filename, role, kind, fileFormat, recursive=False)
            return True
        
    def addGlob(self, glob, role, kind='fashion_unknown', fileFormat='yaml', recursive=True):
        '''Add another glob to this library'''
        g = { 'glob': glob,
              'role': role,
              'kind': kind,
              'recursive': recursive,
              'fileFormat': fileFormat }
        self.globs.extend([g])
        
    def importModels(self):
        '''Import all models in this library into the fashion database.'''
        modelGlobs = [g for g in self.globs if g['role'] == 3]
        # globs are specified relative to library file
        with xformUtil.cd(os.path.dirname(self.filename)):
            for g in modelGlobs:
                recursive = g.get('recursive', True)
                for f in glob.glob(g['glob'], recursive=recursive):
                    fashionPortfolio.importFile(role=g['role'], 
                                                filename=os.path.abspath(f), 
                                                kind=g['kind'],
                                                fileFormat=g['fileFormat'])
        
    def loadXforms(self):
        '''Load all xforms in this library, returned as list.'''
        xfList = []
        xformGlobs = [g for g in self.globs if g['role'] == 4]
        # globs are specified relative to library file
        with xformUtil.cd(os.path.dirname(self.filename)):
            for g in xformGlobs:
                recursive = g.get('recursive', True)
                for f in glob.glob(g['glob'], recursive=recursive):
                    xf = xforms.Xform(f)
                    if xf.exists() and not os.path.isdir(f):
                        xf.load()
                        xfList.extend([xf])
        return xfList
    
    def getDirectories(self, role):
        '''Get the directories which contain model files.'''
        globs = [g for g in self.globs if g['role'] == role]
        # globs are specified relative to library file
        with xformUtil.cd(os.path.dirname(self.filename)):
            dirs = [os.path.abspath(d) for g in globs
                                       for d in glob.glob(g['glob'], recursive=g.get('recursive', True))
                                       if os.path.isdir(d) ]
            logging.debug("getDirectories({0})={1}".format(role, dirs))
            return dirs
        
    def load(self):
        '''Load a library description file.'''
        with open(self.filename, "r") as stream:
            self.globs = ruamel.yaml.load(stream)
            
    def save(self):
        '''Save a library description file.'''
        with open(self.filename, "w") as stream:
            ruamel.yaml.dump(self.globs, stream, default_flow_style = False)



'''
Created on 2016-04-29

Copyright (c) 2016 Bradford Dillman

@author: bdillman

Describe and manipulate xforms, which are python3 modules with specific
requirements.
'''

import importlib.util
import inspect
import logging
import os.path
import traceback

from . import fashionPortfolio
from . import xformUtil



def loadModule(modName, filename):
	'''Load a python module from a file. Python 3.5+'''
	# spec = importlib.util.spec_from_file_location("module.name", "/path/to/file.py")
	spec = importlib.util.spec_from_file_location(modName, filename)
	mod = importlib.util.module_from_spec(spec)
	spec.loader.exec_module(mod)
	return mod

class Xform():
	'''Represents a specific transform operation.'''
	def __init__(self, filename):
		'''Initialize a transform file a Python module filename.'''
		self.filename = filename
		self.modName = os.path.basename(filename).split(".")[0]
		self.isLoaded = False
		
	def load(self):
		'''
		Load the python module for this transform.
		:return: True if loaded
		'''
		if not self.isLoaded:
			try:
				logging.debug("loading xform: {0}".format(self.filename))
				self.spec = importlib.util.spec_from_file_location(self.modName, self.filename)
				self.mod = importlib.util.module_from_spec(self.spec)
				self.spec.loader.exec_module(self.mod)
				try:
					self.outputKinds = self.mod.outputKinds()
				except:
					logging.error("outputKinds error: {0}".format(self.filename))
					traceback.print_exc()
					return False

				func = self.mod.xform
				self.xform_args = inspect.getargspec(func).args
				self.xformInputKinds = set(self.xform_args)
				if hasattr(self.mod, 'police'):
					func = self.mod.police
					self.police_args = inspect.getargspec(func).args
					self.policeInputKinds = set(self.police_args)
					self.inputKinds = self.xformInputKinds | self.policeInputKinds
					self.hasPolice = True
				else:
					self.inputKinds = self.xformInputKinds
					self.hasPolice = False
				self.isLoaded = True
			except AttributeError:
				logging.warning("Failed to load: {0}".format(self.filename))
		return self.isLoaded
	
	def exists(self):
		'''
		Tests for existence of this Xform.
		:returns: true if Xform exists.
		'''
		return os.path.exists(self.filename)
	
	def police(self, **inputs):
		xformUtil.current_xform_name = self.filename
		xformUtil.current_output_kinds = set() # no output for police
		if hasattr(self.mod, 'police'):
			return self.mod.police(**inputs)
	
	def execute(self, **inputs):
		xformUtil.current_xform_name = self.filename
		xformUtil.current_output_kinds = self.outputKinds
		self.mod.xform(**inputs)



class XformPlan(object):
	'''An execution plan for a set of xforms.'''
	def __init__(self, xfCollection = []):
		self.xfSet = set(xfCollection)
		
	def plan(self):
		'''Construction the xform execution plan.'''
		self.goodXforms  = {xf for xf in self.xfSet if xf.isLoaded}
		self.badXforms   = self.xfSet - self.goodXforms
		for xf in self.badXforms:
			logging.warn("bad xform: {0}".format(xf.filename))
		self.xfByName    = {xf.modName:xf for xf in self.goodXforms}
		self.xfNames     = set(self.xfByName)
		
		self.xfOutputs   = {xf.modName:set(xf.outputKinds) for xf in self.goodXforms}
		self.xfInputs    = {xf.modName:set(xf.inputKinds)  for xf in self.goodXforms}
		
		self.allOutputs  = {ok for _, outKinds in self.xfOutputs.items() 
						       for ok in outKinds}
		self.allInputs   = {ik for _, inKinds  in self.xfInputs.items()
						       for ik in inKinds}
		
		self.leafInputs  = self.allInputs  - self.allOutputs
		self.leafoutputs = self.allOutputs - self.allInputs
		self.intermed    = self.allInputs  & self.allOutputs
		
		self.xfByOutput = {}
		for xfName, xfOutKinds in self.xfOutputs.items():
			for outKind in xfOutKinds:
				self.xfByOutput.setdefault(outKind, set([]))
				self.xfByOutput[outKind].add(xfName)
				
		self.xfByInput = {}
		for xfName, xfInKinds in self.xfInputs.items():
			for inKind in xfInKinds:
				self.xfByInput.setdefault(inKind, set([]))
				self.xfByInput[inKind].add(xfName)
			
		self.xfPre = {}
		for xfName in self.xfNames:
			for inp in self.xfInputs[xfName]:
				if inp in self.xfByOutput.keys():
					self.xfPre[xfName] = self.xfByOutput[inp]
		
		self.xfPost = {}
		for xfName in self.xfNames:
			for outp in self.xfOutputs[xfName]:
				if outp in self.xfByInput.keys():
					self.xfPost[xfName] = self.xfByInput[outp]
				
		# now flatten into an exec list
		availInp = self.leafInputs.copy()
		availXforms = self.xfNames.copy()
		self.execList = []
		while(availXforms):
			readyXforms = set()
			for xfName in availXforms:
				if self.xfInputs[xfName] <= availInp:
					readyXforms.add(xfName)
			if readyXforms:
				availXforms = availXforms - readyXforms
				self.execList.extend(readyXforms)
				# readyOutputs might be ready, or only partly complete
				readyOutputs = set() 
				for xfName in readyXforms:
					readyOutputs.update(self.xfOutputs[xfName])
				for outp in readyOutputs:
					for xfName in self.xfByOutput[outp]:
						if xfName in availXforms:
							break
					else:
						availInp.add(outp)
			else:
				break
		if availXforms:
			logging.warn("xform dependency cycle detected")
		for idx, xfName in enumerate(self.execList):
			logging.debug("{0}:{1}".format(idx, xfName))
			
	def execute(self):
		'''Execute all the xforms planned in self.execList.'''
		for xfName in self.execList:
			xf = self.xfByName[xfName]
			
			if xf.hasPolice:
				inputs = {inpKind:fashionPortfolio.getModelFiles(inpKind) 
						  for inpKind in xf.policeInputKinds}
				try:
					if xf.police(**inputs):
						# user verification failed
						logging.error("aborting, police failed: {0}".format(xf.filename))
						return
				except:
					logging.error("aborting, police error: {0}".format(xf.filename))
					traceback.print_exc()
				
			inputs = {inpKind:fashionPortfolio.getModelFiles(inpKind) 
					  for inpKind in xf.xformInputKinds}
			try:
				xf.execute(**inputs)
			except:
				logging.error("aborting, xform error: {0}".format(xf.filename))
				traceback.print_exc()



'''
Created on 2016-04-29

Copyright (c) 2016 Bradford Dillman

@author: bdillman

xform which generates documentation from models and templates.
'''

import logging

from fashion import templates

logging.debug("fashion_doc loaded")

def outputKinds():
    '''Outputs generated by this xform'''
    return [ ]

def police(fashion_gen):
    '''Validate models before xform.'''
    pass

def xform(fashion_gen):
    '''Perform source generation from template.'''
    logging.debug("fashion_doc xform running")
    genModels = [f.loadFile() for f in fashion_gen]
    docModel = {'genModels': genModels}
    args = {'model'       : docModel, 
            'templateFile': 'fashion_build.html', 
            'targetFile'  : 'fashion/build.html'}
    templates.generateFile(**args)



'''
Created on 2016-04-29

Copyright (c) 2016 Bradford Dillman

@author: bdillman

Command line parser for fashion project.
'''

import argparse
import logging
import os
import pathlib
import shutil
import sys

from . import fashionPortfolio
from . import project
from . import templates
from . import xforms
from . import xformUtil
from fashion.project import FASHION_HOME

alwaysYes = False

def query_yes_no(question, default="yes"):
    '''
    Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    '''
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        if alwaysYes:
            sys.stdout.write("y\n")
            return True;
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")

def initCmd(args):
    '''Initialize a new fashion project.'''
    proj = project.Project(args.project)
    if proj.exists():
        print("project already exists")
    else:
        proj.create()

def killCmd(args):
    '''Delete a fashion project.'''
    proj = project.Project(args.project)
    if proj.exists():
        if query_yes_no("Are you sure you want to delete the fashion project", "no"):
            proj.destroy()
    else:
        print("project doesn't exist")

def buildCmd(args):
    '''Perform a full build.'''
    proj = project.findProject(args.project)
    if proj.exists():
        with xformUtil.cd(str(proj.projectPath)):
            proj.init_db()
            fashionPortfolio.clean()
            proj.importModels()
            xfs = proj.loadXforms()
            plan = xforms.XformPlan(xfs)
            plan.plan()
            proj.loadTemplates()
            templates.overwrite = args.force
            plan.execute()
    else:
        print("project doesn't exist")
    
def nabCmd(args):
    '''Nab a file and turn it into a template + xform.'''
    proj = project.findProject(args.project)
    if proj.exists():
        proj.init_db()
        
        targetFile = args.filename
        relTarget = os.path.relpath(targetFile, str(proj.projectPath))

        # create template by copying target file
        templateDir = proj.getLocalTemplateDir()
        templateDest = os.path.join(templateDir, relTarget)
        relTemplate = os.path.relpath(templateDest, str(templateDir))
        if os.path.exists(str(templateDest)):
            print("Template already exists: " + str(templateDest))
            return
        pathlib.Path(os.path.dirname(templateDest)).mkdir(parents=True, exist_ok=True)
        shutil.copyfile(targetFile, str(templateDest))
        print("Created template: " + str(templateDest))
        
        # create a corresponding default xform
        xfName = os.path.basename(os.path.splitext(targetFile)[0])
        if args.name != None:
            xfName = args.name
        xfFilename = xfName + ".py"
        xformDest = os.path.join(proj.getLocalXformDir(), os.path.dirname(relTarget), xfFilename)
        model = { 'name'    : xfName,
                  'template': relTemplate,
                  'target'  : relTarget }
        defaultXformTemplateFile = "defaultXform.py"
        proj.loadTemplates()
        templates.createDefaultFile(model, defaultXformTemplateFile, xformDest)
        print("Created xform: " + str(xformDest))
        
    else:
        print("project doesn't exist")    

def homeCmd(args):
    '''fashion install home'''
    print(str(FASHION_HOME))
    
def infoCmd(args):
    '''fashion install info'''
    print("not implemented")
    
def versionCmd(args):
    '''Version of fashion'''
    print(str(FASHION_HOME))
    print("version: unstable dev")
    
def create_model(args):
    '''Create a new model'''
    proj = project.findProject(args.project)
    if proj.exists():
        proj.init_db()
        targetFile = os.path.join(proj.getLocalModelDir(), args.filename)
        kind = args.kind
        if args.format == None:
            fileFormat = 'yaml'
        else:
            fileFormat = args.format
        lib = proj.getLocalLibrary()
        if lib.addFile(targetFile, 3, kind, fileFormat):
            lib.save()
            if not os.path.exists(targetFile):
                proj.loadTemplates()
                templates.createDefaultFile({}, "defaultModel.yaml", targetFile)
                print("Created new model {0}".format(targetFile))
                return True
            else:
                print("Added existing model {0}".format(targetFile))
                return True
        print("Failed to create model {0}".format(targetFile))
    else:
        print("project doesn't exist")    

def create_xform(args):
    '''Create a new xform'''
    proj = project.findProject(args.project)
    if proj.exists():
        proj.init_db()
        targetFile = os.path.join(proj.getLocalXformDir(), args.filename+'.py')
        lib = proj.getLocalLibrary()
        if lib.addFile(targetFile, 4, fileFormat='python3'):
            lib.save()
            if not os.path.exists(targetFile):
                proj.loadTemplates()
                xfName = os.path.basename(os.path.splitext(targetFile)[0])
                model = { 'name'    : xfName,
                          'template': 'myTemplate',
                          'target'  : 'myTarget' }
                templates.createDefaultFile(model, "defaultXform.py", targetFile)
                print("Created new xform {0}".format(targetFile))
                return True
            else:
                print("Added existing xform {0}".format(targetFile))
                return True
        print("Failed to create xform {0}".format(targetFile))
    else:
        print("project doesn't exist")    

def main():
    '''Parse command from command line args, then delegate.'''
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--project',   help='project directory', nargs=1)
    parser.add_argument('-y', '--alwaysYes', help="answer 'y' to all prompts", action='store_true')
    parser.add_argument('-v', '--verbose',   help="verbose output", action='store_true')
    parser.add_argument('-d', '--debug',     help="debug logging", action='store_true')

    subparsers = parser.add_subparsers(dest='command',
                                       title='commands', 
                                       description='valid commands', 
                                       help='command help')

    initParse = subparsers.add_parser('init', help='initialize a new fashion project in the current directory')

    killParse = subparsers.add_parser('kill', help='delete a fashion project and all contents from the current directory')

    nabParse = subparsers.add_parser('nab', help='nab a file and turn it into a fashionTemplate')
    nabParse.add_argument('filename', help='file to nab')
    nabParse.add_argument('-n', '--name',   help='xform name', nargs=1)

    buildParse = subparsers.add_parser('build', help='full build')
    buildParse.add_argument('-f', '--force', help="force overwrite", action='store_true')
    
    createModelParse = subparsers.add_parser('create-model', help='create a new model')
    createModelParse.add_argument('filename', help='filename for new model')
    createModelParse.add_argument('kind',     help='kind of model to create')
    createModelParse.add_argument('-f', '--format', help="file format", nargs=1)

    createXformParse = subparsers.add_parser('create-xform', help='create a new xform')
    createXformParse.add_argument('filename', help='filename for new xform')

    homeParse = subparsers.add_parser('home', help='report install directory')
    infoParse = subparsers.add_parser('info', help='show info, not implemented')
    versionParse = subparsers.add_parser('version', help='report version information')

    result = parser.parse_args(sys.argv[1:]);
    if result.project == None:
        result.project = os.getcwd()
        
    global alwaysYes        
    if result.alwaysYes:
        alwaysYes = result.alwaysYes

    if result.debug:        
        logging.basicConfig(level=logging.DEBUG)

    logging.debug(result)

    cmds = {
        "init": initCmd,
        "nab": nabCmd,
        "build": buildCmd,
        "kill": killCmd,
        "home": homeCmd,
        "info": infoCmd,
        "version": versionCmd,
        "create-model": create_model,
        "create-xform": create_xform
    }

    try:
        cmds[result.command](result)
    except KeyError:
        if result.command == None:
            parser.print_help()
        else:
            print("unknown command "+result.command);

#
# Main entry point
#
if __name__ == "__main__":
    main(sys.argv[1:])
